from __future__ import annotations

from cleo.helpers import argument
from poetry.console.commands.remove import RemoveCommand

from poetry_types.commands.base import TypesCommand


class RemoveTypesCommand(TypesCommand, RemoveCommand):
    name = "types remove"

    description = "Removes a type stub package from project dependencies."

    arguments = [
        argument(
            "packages", "The packages to remove the type stubs for.", multiple=True
        )
    ]

    help = "<c1>types remove sqlalchemy</c1> removes <c1>types-sqlalchemy</c1> to your project"

    def handle(self) -> int:
        packages = self.argument("packages")
        self.sanitize_types_section()
        packages = set(
            self.find_packages(self.convert_to_type_packages_names(packages))
        )

        existing_packages = self.get_existing_type_packages()

        to_remove = packages.intersection(existing_packages)

        if not to_remove:
            self.line("No type packages to be removed.")
            return 0

        return self.remove_packages(to_remove)



from __future__ import annotations

from cleo.helpers import argument

from poetry_types.commands.base import GROUP_NAME, TypesCommand


class UpdateTypesCommand(TypesCommand):
    name = "types update"

    description = "Adds/Removes type stubs according to pyproject.toml."

    arguments = [
        argument(
            "packages",
            "The packages to update the type stubs for.",
            multiple=True,
            optional=True,
        )
    ]

    help = "<c1>types update</c1> adds/removes type stubs for every dependency in the project"

    def collect_all_packages_not_in_types_section(self) -> dict[str, str]:
        poetry_content = self.pyproject_poetry_content()
        packages = {}
        if "dependencies" in poetry_content:
            packages.update(poetry_content["dependencies"])

        if "group" in poetry_content:
            for group in poetry_content["group"]:
                if group == GROUP_NAME:
                    continue
                group_content = poetry_content["group"][group]
                if "dependencies" in group_content:
                    packages.update(group_content["dependencies"])

        return packages

    def handle(self):
        packages = self.argument("packages")

        whitelist = {}
        self.sanitize_types_section()
        all_packages_not_in_types_section = (
            self.collect_all_packages_not_in_types_section()
        )
        pypi_type_packages = set(
            self.find_packages(
                self.convert_to_type_packages_names(all_packages_not_in_types_section)
            )
        )
        type_section_packages = set(self.get_existing_type_packages())

        to_add = pypi_type_packages.difference(type_section_packages)

        to_remove = {
            package
            for package in type_section_packages.difference(pypi_type_packages)
            if self.is_package_type_package_name(package)
        }

        if to_add:
            requirements = self.install_packages(to_add, True)
            whitelist.update(
                {
                    requirement["name"]: requirement["version"]
                    for requirement in requirements
                }
            )
        if to_remove:
            whitelist.update(self.remove_packages(to_remove, True))

        if packages:
            whitelist.update(
                {name: "*" for name in self.convert_to_type_packages_names(packages)}
            )

        self.installer.whitelist(whitelist)

        self.installer.only_groups([GROUP_NAME])

        self.installer.update(True)

        status = self.installer.run()

        if status == 0 and to_add:
            self.add_constraints_to_toml(requirements)

        return status



from __future__ import annotations

from cleo.helpers import argument

from poetry_types.commands.base import TypesCommand


class AddTypesCommand(TypesCommand):
    name = "types add"

    description = "Adds type stub dependencies for given packages."

    arguments = [
        argument("packages", "The packages to add the typing stubs for", multiple=True)
    ]

    help = (
        "<c1>types add sqlalchemy</c1> adds <c1>types-sqlalchemy</c1> to your project"
    )

    def handle(self) -> int:
        self.sanitize_types_section()
        packages = self.argument("packages")
        packages = self.find_packages(self.convert_to_type_packages_names(packages))
        existing_packages = self.get_existing_type_packages(packages)

        packages = [package for package in packages if package not in existing_packages]

        if not packages:
            self.line("No new type packages to add.")
            return 0

        return self.install_packages(packages)



from __future__ import annotations

import contextlib
import functools
import typing as t

import poetry.core.semver.helpers as semver
import poetry.factory as poetry_factory
import tomlkit
from packaging.utils import canonicalize_name
from poetry.console.commands.init import InitCommand
from poetry.console.commands.installer_command import InstallerCommand

from poetry_types.packages_map import PACKAGES_MAP

if t.TYPE_CHECKING:
    import tomlkit.items as tomlkit_items

GROUP_NAME = "types"


class TypesCommand(InitCommand, InstallerCommand):
    def pyproject_content(self) -> tomlkit.TOMLDocument:
        return self.poetry.file.read()

    def pyproject_poetry_content(self):
        return self.pyproject_content()["tool"]["poetry"]

    @t.overload
    def get_types_section(
        self, create_if_not_exists: t.Literal[False]
    ) -> tomlkit_items.Table | None:
        ...

    @t.overload
    def get_types_section(self) -> tomlkit_items.Table:
        ...

    @t.overload
    def get_types_section(
        self, create_if_not_exists: t.Literal[True]
    ) -> tomlkit_items.Table:
        ...

    def get_types_section(
        self, create_if_not_exists: bool = True
    ) -> tomlkit_items.Table:
        content = self.pyproject_content()

        try:
            return content["tool"]["poetry"]["group"][GROUP_NAME]["dependencies"]
        except KeyError:
            if create_if_not_exists:
                return self.create_types_section()
            return None

    def create_types_section(self):
        content = self.pyproject_content()

        path = ("tool", "poetry", "group", GROUP_NAME, "dependencies")
        section = content
        for n, key in enumerate(path):
            if key not in section:
                section[key] = (
                    tomlkit.table(is_super_table=True)
                    if n < len(path) - 1
                    else tomlkit.table()
                )
            section = section[key]
        self.poetry.file.write(content)
        return section

    def is_package_type_package_name(self, package: str):
        return package.startswith("types-") or package in PACKAGES_MAP.values()

    def convert_to_type_packages_names(self, packages: list[str]):
        return [
            package
            if package.startswith("types-")
            else f"types-{package}"
            if package not in PACKAGES_MAP
            else PACKAGES_MAP[package]
            for package in packages
        ]

    def get_existing_type_packages(
        self, packages: list[str] | None = None, canonicalized=False
    ):
        """Filter packages in type section in pyproject.toml by names"""
        types_section = self.get_types_section(False)

        if types_section is None:
            return []
        if packages:
            return [
                package if not canonicalized else canonicalize_name(package)
                for package in packages
                for key in types_section
                if canonicalize_name(key) == canonicalize_name(package)
            ]

        return [
            package if not canonicalized else canonicalize_name(package)
            for package in types_section
        ]

    def add_constraints_to_toml(self, constraints: list[dict[str, str]]):

        types_section = self.get_types_section()

        content = self.pyproject_content()
        poetry_content = content["tool"]["poetry"]

        for constraint in constraints:
            name = constraint["name"]
            version = constraint["version"]
            types_section[name] = version

        poetry_content["group"][GROUP_NAME]["dependencies"] = types_section

        self.poetry.file.write(content)

    def remove_packages_from_toml(self, packages: list[str]):
        content = self.pyproject_content()
        poetry_content = content["tool"]["poetry"]
        types_section = self.get_types_section()

        if types_section is None:
            return {}

        removed: dict[str, str] = {}
        for package in packages:
            removed[package] = types_section[package]
            del types_section[package]

        if not types_section and GROUP_NAME in poetry_content["group"]:
            del poetry_content["group"][GROUP_NAME]["dependencies"]
        else:
            poetry_content["group"][GROUP_NAME]["dependencies"] = types_section
        self.poetry.file.write(content)
        return removed

    @functools.lru_cache(maxsize=None)
    def find_package(self, package: str):
        try:
            return self._find_best_version_for_package(package)[0]
        except ValueError:
            return None

    def find_packages(self, packages: list[str]):
        found_packages = []
        for package in packages:
            found_package = self.find_package(package)
            if found_package is not None:
                found_packages.append(found_package)
        return found_packages

    def install_packages(self, packages: list[str], prepare_only=False):
        section = self.get_types_section()

        content = self.pyproject_content()

        packages = self.find_packages(packages)

        requirements = self._determine_requirements(packages)

        for _constraint in requirements:
            version = _constraint.get("version")
            if version is not None:
                assert isinstance(version, str)
                semver.parse_constraint(version)

            constraint: dict[str, t.Any] = tomlkit.inline_table()
            for name, value in _constraint.items():
                if name == "name":
                    continue

                constraint[name] = value

            if len(constraint) == 1 and "version" in constraint:
                constraint = constraint["version"]

            constraint_name = _constraint["name"]
            assert isinstance(constraint_name, str)
            section[constraint_name] = constraint

            with contextlib.suppress(ValueError):
                self.poetry.package.dependency_group(GROUP_NAME).remove_dependency(
                    constraint_name
                )

            self.poetry.package.add_dependency(
                poetry_factory.Factory.create_dependency(
                    constraint_name,
                    constraint,
                    groups=[GROUP_NAME],
                    root_dir=self.poetry.file.parent,
                )
            )
        if not prepare_only:
            status = self.installer_action(
                [requirement["name"] for requirement in requirements]
            )

            if status == 0:

                assert isinstance(content, tomlkit.TOMLDocument)

                self.add_constraints_to_toml(requirements)

            return status

        return requirements

    def remove_packages(self, packages: list[str], prepare_only=False):
        removed = self.remove_packages_from_toml(packages)
        group = self.poetry.package.dependency_group(GROUP_NAME)
        for package in packages:
            group.remove_dependency(package)
        if not prepare_only:
            return self.installer_action(packages)
        return removed

    def installer_action(self, packages: list[str]):
        self.poetry.set_locker(
            self.poetry.locker.__class__(
                self.poetry.locker.lock.absolute().as_posix(),
                self.pyproject_poetry_content(),
            )
        )
        self.installer.set_locker(self.poetry.locker)

        self.installer.set_package(self.poetry.package)
        self.installer.update(True)
        self.installer.whitelist(packages)

        return self.installer.run()

    def sanitize_types_section(self):
        types_section = self.get_types_section(False)

        if types_section is None:
            return

        content = self.pyproject_content()

        for package in types_section.copy():
            new_names = self.find_packages([package])
            version = types_section[package]

            if new_names and new_names[0] != package:
                types_section[new_names[0]] = version
                del types_section[package]
            elif not new_names:
                del types_section[package]

        content["tool"]["poetry"]["group"][GROUP_NAME]["dependencies"] = types_section
        self.poetry.file.write(content)

    @classmethod
    def factory(cls):
        return cls()



from poetry.console.application import Application
from poetry.plugins import ApplicationPlugin

from poetry_types.commands import (
    AddTypesCommand,
    RemoveTypesCommand,
    UpdateTypesCommand,
)


class PoetryTypes(ApplicationPlugin):
    def activate(self, app: Application):
        app.command_loader.register_factory("types add", AddTypesCommand.factory)
        app.command_loader.register_factory("types remove", RemoveTypesCommand.factory)
        app.command_loader.register_factory("types update", UpdateTypesCommand.factory)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['poetry_types', 'poetry_types.commands']

package_data = \
{'': ['*']}

install_requires = \
['packaging>=21.3,<22.0', 'poetry>=1.2,<2.0', 'tomlkit>=0.11.4,<0.12.0']

entry_points = \
{'poetry.application.plugin': ['poetry-types = '
                               'poetry_types.poetry_types:PoetryTypes']}

setup_kwargs = {
    'name': 'poetry-types',
    'version': '0.3.5',
    'description': 'A poetry plugin that adds/removes type stubs as dependencies like the mypy --install-types command.',
    'long_description': '# poetry-types\n\n[![PyPI version](https://badge.fury.io/py/poetry-types.svg)](https://badge.fury.io/py/poetry-types)\n[![GitHub license](https://img.shields.io/github/license/jvllmr/poetry-types)](https://github.com/jvllmr/poetry-types/blob/master/LICENSE)\n[![GitHub issues](https://img.shields.io/github/issues/jvllmr/poetry-types)](https://github.com/jvllmr/poetry-types/issues)\n![PyPI - Downloads](https://img.shields.io/pypi/dd/poetry-types)\n![Tests](https://github.com/jvllmr/poetry-types/actions/workflows/main.yml/badge.svg)\n\n## Description\n\nThis is a plugin to poetry for the upcoming poetry 1.2 plugin feature.\nIt installs/removes/updates typing stubs via following commands:\n\n- `poetry types add <package names>`\n- `poetry types remove <package names>`\n- `poetry types update <package names>`\n\n## Usage examples\n\n- `poetry types add SQLAlchemy` adds `types-SQLAlchemy` to your project\n- `poetry types update` adds `types-SQLAlchemy` if `SQLAlchemy` is present, but not `types-SQLAlchemy`\n- `poetry types update` removes `types-SQLAlchemy` if `types-SQLAlchemy` is present, but not `SQLAlchemy`\n\n## Installation\n\nRun `poetry self add poetry-types` for global install or run `poetry add -D poetry-types` to use this plugin with your project.\n\n## Usage with pre-commit\n\n```yaml\n- repo: https://github.com/jvllmr/poetry-types\n  rev: v0.3.5\n  hooks:\n    - id: poetry-types\n```\n\n### poetry-types has to be skipped with pre-commit.ci\n\n```yaml\nci:\n  skip: [poetry-types]\n```\n',
    'author': 'Jan Vollmer',
    'author_email': 'jan@vllmr.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jvllmr/poetry-types',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
